@@ -62,6 +62,8 @@ typedef struct _WnckTaskClass   WnckTaskClass;
 #define DEFAULT_WIDTH 1
 #define DEFAULT_HEIGHT 48
 
+#define MAX_WIDTH_CHARS 250
+
 #define N_SCREEN_CONNECTIONS 5
 
 #define POINT_IN_RECT(xcoord, ycoord, rect) \
@@ -116,6 +118,8 @@ struct _WnckTask
   guint was_active : 1;      /* used to fixup activation behavior */ 
 
   guint button_activate;
+
+  guint32 dnd_timestamp;
 };
 
 struct _WnckTaskClass
@@ -240,7 +244,8 @@ static void     wnck_tasklist_connect_window           (WnckTasklist *tasklist,
 static void     wnck_tasklist_change_active_task       (WnckTasklist *tasklist,
 							WnckTask *active_task);
 static gboolean wnck_tasklist_change_active_timeout    (gpointer data);
-static void     wnck_tasklist_activate_task_window     (WnckTask *task);
+static void     wnck_tasklist_activate_task_window     (WnckTask *task,
+                                                        guint32   timestamp);
 
 static void     wnck_tasklist_update_icon_geometries   (WnckTasklist *tasklist,
 							GList        *visible_tasks);
@@ -1154,8 +1159,8 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   while (l != NULL)
     {
       WnckTask *task = WNCK_TASK (l->data);
-      int col = i % n_cols;
-      int row = i / n_cols;
+      int row = i % n_rows;
+      int col = i / n_rows;
       
       child_allocation.x = total_width*col / n_cols;
       child_allocation.y = allocation->height*row / n_rows;
@@ -1788,7 +1793,8 @@ wnck_tasklist_window_changed_geometry (WnckWindow   *window,
    * the tasklist itself possibly changed monitor.
    */
   monitor_changed = FALSE;
-  if (wnck_window_get_state (window) & WNCK_WINDOW_STATE_SKIP_TASKLIST &&
+  if (tasklist->priv->monitor_num != -1 &&
+      (wnck_window_get_state (window) & WNCK_WINDOW_STATE_SKIP_TASKLIST) &&
       GTK_WIDGET (tasklist)->window != NULL)
     {
       /* Do the extra check only if there is a suspect of a monitor change (= this window is off monitor) */
@@ -1872,7 +1878,9 @@ wnck_task_position_menu (GtkMenu   *menu,
   GtkRequisition requisition;
   gint menu_xpos;
   gint menu_ypos;
-
+  gint pointer_x;
+  gint pointer_y;
+  
   gtk_widget_size_request (GTK_WIDGET (menu), &requisition);
 
   gdk_window_get_origin (widget->window, &menu_xpos, &menu_ypos);
@@ -1885,6 +1893,10 @@ wnck_task_position_menu (GtkMenu   *menu,
   else
     menu_ypos += widget->allocation.height;
 
+  gtk_widget_get_pointer (widget, &pointer_x, &pointer_y);
+  if (requisition.width < pointer_x)
+    menu_xpos += MIN (pointer_x, widget->allocation.width - requisition.width);
+
   *x = menu_xpos;
   *y = menu_ypos;
   *push_in = TRUE;
@@ -1908,11 +1920,15 @@ wnck_task_menu_activated (GtkMenuItem *menu_item,
 {
   WnckTask *task = WNCK_TASK (data);
 
-  wnck_tasklist_activate_task_window (task);
+  /* This is an "activate" callback function so gtk_get_current_event_time()
+   * will suffice.
+   */
+  wnck_tasklist_activate_task_window (task, gtk_get_current_event_time ());
 }
 
 static void
-wnck_tasklist_activate_task_window     (WnckTask *task)
+wnck_tasklist_activate_task_window (WnckTask *task,
+                                    guint32   timestamp)
 {
   WnckTasklist *tasklist;
   WnckWindowState state;
@@ -1933,10 +1949,10 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
       window_ws = wnck_window_get_workspace (task->window);
       if (window_ws &&
           active_ws != window_ws &&
-	  !tasklist->priv->switch_workspace_on_unminimize)
-	wnck_workspace_activate (window_ws);
-	  
-      wnck_window_activate_transient (task->window);
+          tasklist->priv->switch_workspace_on_unminimize)
+        wnck_window_move_to_workspace (task->window, active_ws);
+
+      wnck_window_activate_transient (task->window, timestamp);
     }
   else
     {
@@ -1948,14 +1964,8 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
 	}
       else
 	{
-          WnckWorkspace *window_ws;
-          
-          window_ws = wnck_window_get_workspace (task->window);
-          if (window_ws)
-            wnck_workspace_activate (window_ws);
-
-	  wnck_window_activate_transient (task->window);
-	}
+          wnck_window_activate_transient (task->window, timestamp);
+        }
     }
   
 
@@ -1968,6 +1978,56 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
   wnck_tasklist_change_active_task (tasklist, task);
 }
 
+static void 
+wnck_task_close_all (GtkMenuItem *menu_item,
+ 		     gpointer     data)
+{
+  WnckTask *task = WNCK_TASK (data);
+  GList *l;
+
+  l = task->windows;
+  while (l)
+    {
+      WnckTask *child = WNCK_TASK (l->data);
+      wnck_window_close (child->window, gtk_get_current_event_time ());
+      l = l->next;
+    }
+}
+
+static void
+wnck_task_unminimize_all (GtkMenuItem *menu_item,
+		          gpointer     data)
+{
+  WnckTask *task = WNCK_TASK (data);
+  GList *l;
+
+  l = task->windows;
+  while (l)
+    {
+      WnckTask *child = WNCK_TASK (l->data);
+      /* This is inside an activate callback, so gtk_get_current_event_time()
+       * will work.
+       */
+      wnck_window_unminimize (child->window, gtk_get_current_event_time ());
+      l = l->next;
+    }
+}
+
+static void 
+wnck_task_minimize_all (GtkMenuItem *menu_item,
+  		        gpointer     data)
+{
+  WnckTask *task = WNCK_TASK (data);
+  GList *l;
+
+  l = task->windows;
+  while (l)
+    {
+      WnckTask *child = WNCK_TASK (l->data);
+      wnck_window_minimize (child->window);
+      l = l->next;
+    }
+}
 
 static void
 wnck_task_popup_menu (WnckTask *task,
@@ -2021,20 +2081,64 @@ wnck_task_popup_menu (WnckTask *task,
       
       gtk_widget_show (menu_item);
 
-       if (action_submenu)
-         gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),
-                                    wnck_create_window_action_menu (win_task->window));
-       else
-         g_signal_connect_object (G_OBJECT (menu_item), "activate",
-                                  G_CALLBACK (wnck_task_menu_activated),
-                                  G_OBJECT (win_task),
-                                  0);
+      if (action_submenu)
+        gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item),
+                                   wnck_create_window_action_menu (win_task->window));
+      else
+        g_signal_connect_object (G_OBJECT (menu_item), "activate",
+                                 G_CALLBACK (wnck_task_menu_activated),
+                                 G_OBJECT (win_task),
+                                 0);
       
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
       
       l = l->next;
     }
 
+  /* In case of Right click, show Close All, Minimize All, Unminimize All*/
+  if (action_submenu) 
+    {
+      GtkWidget *separator;
+      GtkWidget *image;
+		
+      separator = gtk_separator_menu_item_new ();
+      gtk_widget_show (separator);
+      gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), separator);
+
+      menu_item = gtk_image_menu_item_new_with_mnemonic(_("_Close All"));
+      image = gtk_image_new_from_stock (WNCK_STOCK_DELETE, GTK_ICON_SIZE_MENU);
+      gtk_widget_show (image);
+      gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);    
+      gtk_widget_show (menu_item);
+      gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), menu_item);
+      g_signal_connect_object (G_OBJECT (menu_item), "activate",
+			       G_CALLBACK (wnck_task_close_all),
+			       G_OBJECT (task),
+			       0);
+
+      separator = gtk_separator_menu_item_new ();
+      gtk_widget_show (separator);
+      gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), separator);
+	    
+      menu_item = gtk_image_menu_item_new_with_mnemonic (_("_Minimize All"));
+      image = gtk_image_new_from_stock (WNCK_STOCK_MINIMIZE, GTK_ICON_SIZE_MENU);
+      gtk_widget_show (image);
+      gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);  	
+      gtk_widget_show (menu_item);
+      gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), menu_item);
+      g_signal_connect_object (G_OBJECT (menu_item), "activate",
+	    		       G_CALLBACK (wnck_task_minimize_all),
+			       G_OBJECT (task),
+			       0);
+		
+      menu_item =  gtk_image_menu_item_new_with_mnemonic (_("_Unminimize All"));
+      gtk_widget_show (menu_item);
+      gtk_menu_shell_prepend (GTK_MENU_SHELL (menu), menu_item);			
+      g_signal_connect_object (G_OBJECT (menu_item), "activate",
+  			       G_CALLBACK (wnck_task_unminimize_all),
+			       G_OBJECT (task),
+			       0);							
+    }
   gtk_menu_set_screen (GTK_MENU (menu),
 		       _wnck_screen_get_gdk_screen (task->tasklist->priv->screen));
   
@@ -2067,8 +2171,11 @@ wnck_task_button_toggled (GtkButton *button,
     case WNCK_TASK_WINDOW:
       if (task->window == NULL)
 	return;
-      
-      wnck_tasklist_activate_task_window (task);
+
+      /* This should only be called by clicking on the task button, so
+       * gtk_get_current_event_time() should be fine here...
+       */
+      wnck_tasklist_activate_task_window (task, gtk_get_current_event_time ());
       break;
     case WNCK_TASK_STARTUP_SEQUENCE:
       break;
@@ -2278,7 +2385,7 @@ wnck_task_get_demands_attention (WnckTask *task)
 	{
 	  win_task = WNCK_TASK (l->data);
 
-	  if (wnck_window_get_state (win_task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION)
+	  if (wnck_window_or_transient_demands_attention (win_task->window))
 	    {
 	      demands_attention = TRUE;
 	      break;
@@ -2290,7 +2397,7 @@ wnck_task_get_demands_attention (WnckTask *task)
 
     case WNCK_TASK_WINDOW:
       demands_attention =
-	wnck_window_get_state (task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION;
+	wnck_window_or_transient_demands_attention (task->window);
       break;
 
     case WNCK_TASK_STARTUP_SEQUENCE:
@@ -2412,7 +2519,9 @@ wnck_task_motion_timeout (gpointer data)
 
   task->button_activate = 0;
 
-  wnck_window_activate_transient (task->window);
+  wnck_window_activate_transient (task->window, task->dnd_timestamp);
+
+  task->dnd_timestamp = 0;
 
   return FALSE;
 }
@@ -2438,14 +2547,14 @@ wnck_task_drag_motion (GtkWidget          *widget,
 		       guint               time,
 		       WnckTask            *task)
 {
+  task->dnd_timestamp = time;
+  if (task->button_activate == 0 && task->type == WNCK_TASK_WINDOW)
+      task->button_activate = g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
+                                             wnck_task_motion_timeout,
+                                             task);
+  gdk_drag_status (context,0,time);
 
-   if (task->button_activate == 0 && task->type == WNCK_TASK_WINDOW)
-       task->button_activate = g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
-                                              wnck_task_motion_timeout,
-                                              task);
-   gdk_drag_status (context,0,time);
-
-   return TRUE;
+  return TRUE;
 }
 
 static gboolean
@@ -2513,11 +2622,10 @@ wnck_task_button_press_event (GtkWidget	      *widget,
   return FALSE;
 }
 
-
 static void
 wnck_task_create_widgets (WnckTask *task)
 {
-  GtkWidget *table;
+  GtkWidget *hbox;
   GdkPixbuf *pixbuf;
   char *text;
   static GQuark disable_sound_quark = 0;
@@ -2540,7 +2648,7 @@ wnck_task_create_widgets (WnckTask *task)
 
   gtk_drag_dest_set (GTK_WIDGET(task->button), 0, NULL, 0, 0);
 
-  table = gtk_table_new (1, 2, FALSE);
+  hbox = gtk_hbox_new (FALSE, 0);
 
   pixbuf = wnck_task_get_icon (task);
   if (pixbuf)
@@ -2558,26 +2666,17 @@ wnck_task_create_widgets (WnckTask *task)
   gtk_misc_set_alignment (GTK_MISC (task->label), 0.0, 0.5);
   gtk_label_set_ellipsize (GTK_LABEL (task->label),
                           PANGO_ELLIPSIZE_END);
+  gtk_label_set_max_width_chars (GTK_LABEL (task->label), MAX_WIDTH_CHARS);
 
   if (wnck_task_get_demands_attention (task))
     make_gtk_label_bold ((GTK_LABEL (task->label)));
   gtk_widget_show (task->label);
 
-  gtk_table_attach (GTK_TABLE (table),
-		    task->image,
-		    0, 1,
-		    0, 1,
-		    0, GTK_EXPAND,
-		    4, 0);
-  gtk_table_attach (GTK_TABLE (table),
-		    task->label,
-		    1, 2,
-		    0, 1,
-		    GTK_FILL | GTK_EXPAND, GTK_EXPAND,		    
-		    2, 0);
+  gtk_box_pack_start (GTK_BOX (hbox), task->image, FALSE, FALSE, 4);
+  gtk_box_pack_start (GTK_BOX (hbox), task->label, TRUE, TRUE, 2);
 
-  gtk_container_add (GTK_CONTAINER (task->button), table);
-  gtk_widget_show (table);
+  gtk_container_add (GTK_CONTAINER (task->button), hbox);
+  gtk_widget_show (hbox);
   
   gtk_tooltips_set_tip (task->tasklist->priv->tooltips, task->button, text, NULL);
   g_free (text);
@@ -2682,11 +2781,10 @@ wnck_task_compare (gconstpointer  a,
 {
   WnckTask *task1 = WNCK_TASK (a);
   WnckTask *task2 = WNCK_TASK (b);
-  gulong xid1_1, xid1_2;
-  gulong xid2_1, xid2_2;
-  
-  xid1_1 = xid1_2 = xid2_1 = xid2_2 = 0;   /* silence compiler */
-  
+  gint pos1, pos2;
+
+  pos1 = pos2 = 0;  /* silence the compiler */
+
   switch (task1->type)
     {
     case WNCK_TASK_CLASS_GROUP:
@@ -2696,11 +2794,10 @@ wnck_task_compare (gconstpointer  a,
 	return -1; /* Sort groups before everything else */
 
     case WNCK_TASK_WINDOW:
-      xid1_1 = wnck_window_get_group_leader (task1->window);
-      xid1_2 = wnck_window_get_xid (task1->window);
+      pos1 = wnck_window_get_sort_order (task1->window);
       break;
     case WNCK_TASK_STARTUP_SEQUENCE:
-      xid1_1 = xid1_2 = G_MAXULONG;
+      pos1 = G_MAXINT;
       break;
     }
 
@@ -2713,20 +2810,20 @@ wnck_task_compare (gconstpointer  a,
 	return 1; /* Sort groups before everything else */
 
     case WNCK_TASK_WINDOW:
-      xid2_1 = wnck_window_get_group_leader (task2->window);
-      xid2_2 = wnck_window_get_xid (task2->window);
+      pos2 = wnck_window_get_sort_order (task2->window);
       break;
     case WNCK_TASK_STARTUP_SEQUENCE:
-      xid2_1 = xid2_2 = G_MAXULONG;
+      pos2 = G_MAXINT;
       break;
     }
       
-  if ((xid1_1 < xid2_1) || ((xid1_1 == xid2_1) && (xid1_2 < xid2_2)))
+  if (pos1 < pos2)
     return -1;
-  else if ((xid1_1 == xid2_1) && (xid1_2 == xid2_2))
-    return 0;
-  else
+  else if (pos1 > pos2)
     return 1;
+  else
+    return 0; /* should only happen if there's multiple processes being
+               * started, and then who cares about sort order... */
 }
 
 static void
